#!/usr/bin/env python

import os
import random
import string
from subprocess import call

import numpy as np

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.datasets import dump_svmlight_file


FESTPATH = "/data/malapp_analysis/lib/fest/"


def id_generator(size=6, chars=string.ascii_lowercase + string.digits):
    return ''.join(random.choice(chars) for x in range(size))


class FEST(BaseEstimator, ClassifierMixin):
    """
    Simple Python Wrapper for FEST (https://github.com/fest/fest)
    Supports only binary classification.
    """

    def __init__(self, _depth=1000, _num_tree=100, var_imp=False):
        self.model = id_generator(10)
        self.alg_id = 3
        self.tree_depth = _depth
        self.tree_num = _num_tree
        self.compute_var_imp = var_imp

    def __init__(self, ID, _alg_id, _depth=1000, _num_tree=100, var_imp=False):
        self.model = ID
        self.alg_id = _alg_id
        self.tree_depth = _depth
        self.tree_num = _num_tree
        self.compute_var_imp = var_imp

    def __del__(self):
        # call(["rm", self.model_filename])
        pass

    def set_tree_num(self, _num):
        self.tree_num = _num

    def set_tree_depth(self, _depth):
        self.tree_depth = _depth

    def fit(self, X, y):
        self.train_filename = self.model + "_data.tmp"
        self.model_filename = self.model + "_model.tmp"

        with open(self.train_filename, 'wb') as f:
            dump_svmlight_file(X, y, f)

        cmd_list = [os.path.join(FESTPATH, "festlearn"),
                    "-c", str(self.alg_id),
                    "-d", str(self.tree_depth),
                    "-t", str(self.tree_num)
                    ]
        if (self.compute_var_imp):
            cmd_list.append("-i")

        cmd_list.extend([self.train_filename, self.model_filename])
        call(cmd_list)
        call(["rm", self.train_filename])

    def predict_proba(self, X):
        self.test_filename = self.model + "_test.tmp"
        self.pred_filename = self.model + "_pred.tmp"

        y = np.array([-55] * X.shape[0])

        with open(self.test_filename, 'wb') as f:
            dump_svmlight_file(X, y, f)

        call([os.path.join(FESTPATH, "festclassify"), self.test_filename,
              self.model_filename, self.pred_filename])

        with open(self.pred_filename, 'rb') as f:
            y_pred = np.array([[-float(line), float(line)] for line in f])

        call(["rm", self.test_filename])
        # call(["rm", self.pred_filename])

        return np.array(y_pred)

    def predict(self, X):
        return self.predict_proba(X)[:, 1]
