#!/usr/bin/env python
# -*- coding:utf-8 -*-

import os
import sys
import sklearn.datasets as skd
import cPickle as pickle
from pprint import pprint
from optparse import OptionParser
from sklearn.tree import DecisionTreeClassifier
from plot_roc import roc_curve
from compute_perf_stat import stat_compute, PerfStat, Performance
from compute_perf_stat import plot_perf_stat


def do_skdt_train_predict(train_y, train_x, test_y, test_x, k):
    criterion = options.criterion
    dtc = DecisionTreeClassifier(criterion=criterion,
                                 splitter='best',
                                 max_features='auto',
                                 max_depth=options.tree_depth
                                 # class_weight={'-1': 1, '1': 4}
                                 )
    print "decision tree initialized"

    dtc.fit(train_x, train_y)
    print "decision tree fitted"

    deci_values = dtc.predict_proba(test_x)
    print "dtc predicted, deci_values.shape: "
    pprint(deci_values.shape)

    save_model(dtc, 'model_k-%d.dat' % k)

    posi_idx = 0 if dtc.classes_[0] == 1 else 1
    print posi_idx
    pprint(dtc.classes_)

    pred_labels = [dtc.classes_[0] if x[0] > x[1] else dtc.classes_[1]
                   for x in deci_values]
    stat = stat_compute(test_y, pred_labels)
    return test_y, deci_values[:, posi_idx], stat


def save_model(model, fname):
    with open(fname, 'w') as f:
        pickle.dump(model, f, True)


def save_fpr_tpr_testy(deci, test_y, iternum):
    fprs, tprs, auc = roc_curve(deci, test_y)
    with open('FPRs-TPRs-iter-%d.txt' % (iternum), 'w') as f:
        for i in xrange(len(fprs)):
            print >>f, fprs[i], tprs[i]

    with open('actual_label_%d.txt' % (iternum), 'w') as f:
        for i in xrange(len(test_y)):
            print >>f, int(test_y[i])


def save_temp_result(outdir, k,  perf):
    file_name_suffix = "_k-%d" % (k)
    deci_values_file = os.path.join(
        outdir, "deci_values%s.txt" % (file_name_suffix))

    with open(deci_values_file, 'w') as f:
        for v in perf.deci_value[k]:
            print >>f, v


def plot_stat_and_roc(perf_total, nfold, out_dir='.'):
    for i in range(0, 7):
        plot_perf_stat(perf_total,
                       None,
                       [x for x in range(nfold)],
                       'iter',
                       i,
                       'single',
                       out_dir)
    plot_perf_stat(perf_total, None, None, None, None, 'roc', out_dir)


def do_cross_validation_in_norm_dir(norm_dir, out_dir='.'):
    nfold = options.nfold
    perf = Performance('SK-DT',
                       'decisiontree',
                       [],
                       [],
                       [],
                       nfold,
                       '')
    basename = os.path.basename(os.path.realpath(norm_dir))

    for k in range(nfold):
        print 'sklearn decision tree iter: %d' % k
        train_file = os.path.join(norm_dir,
                                  '{}-train-{}-scaled'.format(basename, k))
        test_file = os.path.join(norm_dir,
                                 '{}-test-{}-scaled'.format(basename, k))

        train_x, train_y = skd.load_svmlight_file(train_file,
                                                  n_features=options.dimension)
        print "train file loaded"
        test_x, test_y = skd.load_svmlight_file(test_file,
                                                n_features=options.dimension)
        print "test file loaded"

        test_y, deci_values, evals = do_skdt_train_predict(train_y, train_x,
                                                           test_y, test_x, k)

        perf.actual_data.append(test_y)
        perf.deci_value.append(deci_values)
        perf.perf_stat.append(PerfStat(evals))
        save_fpr_tpr_testy(deci_values, test_y, k)
        save_temp_result(out_dir, k, perf)

    perf.average()
    plot_stat_and_roc([perf], nfold)
    perf.save('.')
    perf.save_each_stat('.')


def main(options, arguments):
    if options.normalized_dir:
        do_cross_validation_in_norm_dir(options.normalized_dir)
        exit()


if __name__ == "__main__":
    option_1 = {'name': ('-c', '--criterion'), 'help': 'gini or entropy',
                'default': 'gini', 'nargs': 1}
    option_2 = {'name': ('-v', '--nfold'), 'type': 'int',
                'help': 'fold number of cross validation',
                'default': '5', 'nargs': 1}
    option_3 = {'name': ('-d', '--dimension'), 'help': 'number of features',
                'type': 'int', 'nargs': 1}
    option_4 = {'name': ('-o', '--out-dir'), 'default': './',
                'help': 'output directory', 'nargs': 1}
    option_5 = {'name': ('-D', '--tree-depth'),
                'help': 'depth of decision trees',
                'type': 'int', 'default': '1000', 'nargs': 1}
    option_6 = {'name': ('-n', '--normalized-dir'),
                'help': 'load normalized data in this dir',
                'nargs': 1}

    options = [option_1, option_2, option_3,
               option_4, option_5, option_6]

    parser = OptionParser()
    for option in options:
        param = option['name']
        del option['name']
        parser.add_option(*param, **option)

    options, arguments = parser.parse_args()
    sys.argv[:] = arguments

    main(options, arguments)
