#!/usr/bin/env python
# standardize the dataset.
# i.e. for each feature, subtracting the mean and dividing by the
# standard deviation.

import sys
import numpy as np
from svmutil import *


def standardize(x_train, x_test):
    xz_train, mean, sigma = zscore(x_train)
    xz_test = zscore(x_test, mean, sigma, True)

    return xz_train, xz_test


def toarray(X, N=None):
    if N is None:
        # training data
        xarr = []
        N = 0
        for xi in X:
            nextj = 0
            xarri = []
            for k in sorted(xi):
                xarri.extend([0.0 for j in range(nextj, k-1)])
                xarri.append(xi[k])
                nextj = k
            if nextj > N:
                N = nextj
            xarr += [xarri]

        for xarri in xarr:
            if len(xarri) < N:
                l = len(xarri)
                xarri.extend([0.0 for j in range(l, N)])
        return xarr, N

    else:
        # testing data
        xarr = []
        for xi in X:
            xarri = [0.0 for i in range(N)]
            for k in xi:
                if k <= N:
                    xarri[k-1] = xi[k]
                else:
                    break
            xarr += [xarri]
        return xarr


def zscore(X, mean=None, sigma=None, testing=False):
    if testing:
        # testing data
        if mean is None or sigma is None or (len(mean) != len(sigma)):
            return None, None, None
        N = len(mean)
        xarr = toarray(X, N)
    else:
        # training data
        xarr, N = toarray(X)
        mean = np.mean(xarr, axis=0)
        sigma = np.std(xarr, axis=0)

    for i in range(len(xarr)):
        for j in range(len(xarr[i])):
            if sigma[j] == 0:
                xarr[i][j] = xarr[i][j] - mean[j]
            else:
                xarr[i][j] = (xarr[i][j] - mean[j]) / sigma[j]
    if testing:
        return xarr
    else:
        return xarr, mean, sigma


def dump_svmlight_file(y, x, filename):
    with open(filename, 'w') as f:
        for i in range(len(x)):
            print >>f, "%+d" % (y[i]),
            for j in range(len(x[i])):
                print >>f, "%d:%f" % (j+1, x[i][j]),
            print >>f, ''


def main():
    train_file = sys.argv[1]
    test_file = sys.argv[2]
    Y_train, X_train = svm_read_problem(train_file)
    Y_test, X_test = svm_read_problem(test_file)
#    x_zscore, mean, sigma = zscore(X_train)
#    xtest_zscore = zscore(X_test, mean, sigma, True)
    x_zscore, xtest_zscore = standardize(X_train, X_test)
    dump_svmlight_file(Y_train, x_zscore, 'traindata.txt')
    dump_svmlight_file(Y_test, xtest_zscore, 'testdata.txt')


if __name__ == "__main__":
    main()
